class AdfotgError(Exception):
    '''Error base for all errors generated by this app.'''
    pass


class ActionError(AdfotgError):
    '''This error denotes an invalid operation caused by the user.

    The API server translates it to code 400.

    '''
    pass
